# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
This module contains functions that work on ``Optional`` values. In contrast to
other approaches, this does not wrap values into a dedicated type, but works
on normal Python values, which are of type ``Optional[T]``.

Thus, some functions are implemented identically but have different type
signatures. For example, both ``map`` and ``and_then`` both just apply a
function to a value if it is not ``None``, but the result of `map` is ``T``
and the result of ``and_then`` is ``Optional[T]``.

Each function is implemented twice, as a simple function and as a method on
``maybe.Maybe``::

    maybe.Maybe(1).map(fn)

    maybe.map(1, fn)

The names are taken from Rust, see:
https://doc.rust-lang.org/stable/std/option/enum.Option.html

.. note::
    The argument order for ``map_or`` and ``map_or_else`` is reversed
    compared to their Rust counterparts.

    ``do`` is not implemented in Rust but mimics ``toolz.do`` instead.

"""

from dataclasses import dataclass
from typing import Callable, Generic, Optional, TypeVar

T = TypeVar("T")
U = TypeVar("U")


def expect(val: Optional[T], msg: str) -> T:
    """
    Ensure that ``val`` is not ``None``, raises a ``ValueError`` using ``msg``
    otherwise.

    >>> expect(1, "My message")
    1
    >>> expect(None, "My message")
    Traceback (most recent call last):
        ...
    ValueError: My message

    """
    if val is None:
        raise ValueError(msg)

    return val


def do(val: Optional[T], fn: Callable[[T], U]) -> Optional[T]:
    """
    Apply ``fn`` to ``val`` then return ``val``, if ``val`` is not ``None``.

    >>> do("a", print)
    a
    'a'
    >>> do(None, print)

    """
    if val is not None:
        fn(val)

    return val


def map(val: Optional[T], fn: Callable[[T], U]) -> Optional[U]:
    """
    Apply ``fn`` to ``val`` if ``val`` is not ``None``.

    >>> map(1, lambda x: x + 1)
    2
    >>> map(None, lambda x: x + 1)

    """
    return map_or(val, fn, None)


def map_or(val: Optional[T], fn: Callable[[T], U], default: U) -> U:
    """
    Apply ``fn`` to ``val`` if ``val`` is not ``None`` and return the result.
    In case of ``None`` the provided ``default`` is returned instead.

    This is similar to calling ``map`` and ``unwrap_or`` in succession.

    >>> map_or(["x"], len, 0)
    1
    >>> map_or(None, len, 0)
    0

    """
    if val is None:
        return default

    return fn(val)


def map_or_else(
    val: Optional[T],
    fn: Callable[[T], U],
    factory: Callable[[], U],
) -> U:
    """
    Similar to ``map_or``, except that the returned value is lazily evaluated.

    This is similar to calling ``map`` and ``unwrap_or_else`` in succession.

    >>> map_or_else(1, lambda n: [n], list)
    [1]
    >>> map_or_else(None, lambda n: [n], list)
    []

    """
    if val is None:
        return factory()

    return fn(val)


def unwrap(val: Optional[T]) -> T:
    """
    Assert that the value is not ``None``.

    >>> unwrap(1)
    1
    >>> unwrap(None)
    Traceback (most recent call last):
        ...
    ValueError: Trying to unwrap `None` value.

    """
    return expect(val, "Trying to unwrap `None` value.")


def unwrap_or(val: Optional[T], default: T) -> T:
    """
    Get ``val`` if it is not ``None``, or ``default`` otherwise.

    >>> unwrap_or(1, 2)
    1
    >>> unwrap_or(None, 2)
    2

    """
    if val is None:
        return default

    return val


def unwrap_or_else(val: Optional[T], factory: Callable[[], T]) -> T:
    """
    Get ``val`` if it is not ``None``, or invoke ``factory`` to get a fallback.

    >>> unwrap_or_else([1, 2, 3], list)
    [1, 2, 3]
    >>> unwrap_or_else(None, list)
    []

    """
    if val is None:
        return factory()

    return val


def and_(val: Optional[T], other: Optional[U]) -> Optional[U]:
    """
    Like ``a and b`` in Python, except only considering ``None``.

    This is implement identical to ``unwrap_or``, but different with respect to
    types.

    >>> and_(1, 2)
    2
    >>> and_(1, None)
    >>> and_(None, 2)

    """
    if val is None:
        return None

    return other


def and_then(val: Optional[T], fn: Callable[[T], Optional[U]]) -> Optional[U]:
    """
    Apply ``fn`` to ``val`` if it is not ``None`` and return the result.

    In contrast to ``map``, ``fn`` always returns an ``Optional``, which is
    consequently flattened.

    >>> and_then([42], lambda xs: xs[0] if xs else None)
    42
    >>> and_then(None, lambda xs: xs[0] if xs else None)
    """
    return flatten(map(val, fn))


def or_(val: Optional[T], default: Optional[T]) -> Optional[T]:
    """
    Like ``a or b`` in Python, except only considering ``None``.

    >>> or_(1, 2)
    1
    >>> or_(1, None)
    1
    >>> or_(None, 2)
    2
    """
    if val is None:
        return default

    return val


def or_else(
    val: Optional[T], factory: Callable[[], Optional[T]]
) -> Optional[T]:
    """
    Like ``unwrap_or_else``, except that it returns an optional value.

    >>> or_else([42], list)
    [42]
    >>> or_else(None, list)
    []
    """
    return unwrap_or_else(val, factory)  # type: ignore


def contains(val: Optional[T], other: U) -> bool:
    """
    Check if ``val`` equals ``other``, always return ``False`` if ``val`` is
    ``None``.

    >>> contains(1, 1)
    True
    >>> contains(1, 2)
    False
    >>> contains(None, 3)
    False
    """
    if val is None:
        return False

    return val == other


def filter(val: Optional[T], pred: Callable[[T], bool]) -> Optional[T]:
    """
    Return ``None`` if ``val`` is ``None`` or if ``pred(val)`` does not return
    ``True``, otherwise return ``val``.

    >>> is_even = lambda n: n % 2 == 0
    >>> filter(1, is_even)
    >>> filter(2, is_even)
    2
    >>> filter(None, is_even)
    """
    if val is None or not pred(val):
        return None

    return val


def xor(val: Optional[T], other: Optional[T]) -> Optional[T]:
    """
    Return either ``val`` or ``other`` if the other is ``None``. Also return
    ``None`` if both are not ``None``.

    >>> xor(1, None)
    1
    >>> xor(None, 2)
    2
    >>> xor(1, 2)
    >>> xor(None, None)

    """

    if val is None:
        return other

    if other is None:
        return val

    return None


def flatten(val: Optional[Optional[T]]) -> Optional[T]:
    """Flatten nested optional value.

    Note: This just returns the value, but changes the type from
    ``Optional[Optional[T]]`` to ``Optional[T].``
    """

    return val  # type: ignore


@dataclass
class Maybe(Generic[T]):
    val: Optional[T]

    def expect(self, msg: str) -> T:
        """
        Ensure that ``val`` is not ``None``, raises a ``ValueError`` using
        ``msg`` otherwise.

        >>> Maybe(1).expect("My message")
        1
        >>> Maybe(None).expect("My message")
        Traceback (most recent call last):
            ...
        ValueError: My message

        """
        return expect(self.val, msg)

    def do(self, fn: Callable[[T], U]) -> Optional[T]:
        """
        Apply ``fn`` to ``val`` then return ``val``, if ``val`` is not
        ``None``.

        >>> Maybe("a").do(print)
        a
        'a'
        >>> Maybe(None).do(print)

        """
        return do(self.val, fn)

    def map(self, fn: Callable[[T], U]) -> Optional[U]:
        """
        Apply ``fn`` to ``val`` if ``val`` is not ``None``.

        >>> Maybe(1).map(lambda x: x + 1)
        2
        >>> Maybe(None).map(lambda x: x + 1)

        """
        return map(self.val, fn)

    def map_or(self, fn: Callable[[T], U], default: U) -> U:
        """
        Apply ``fn`` to ``val`` if ``val`` is not ``None`` and return the
        result. In case of ``None`` the provided ``default`` is returned
        instead.

        This is similar to calling ``map`` and ``unwrap_or`` in succession.

        >>> Maybe(["x"]).map_or(len, 0)
        1
        >>> Maybe(None).map_or(len, 0)
        0

        """
        return map_or(self.val, fn, default)

    def map_or_else(
        self,
        fn: Callable[[T], U],
        factory: Callable[[], U],
    ) -> U:
        """
        Similar to ``map_or``, except that the returned value is lazily
        evaluated.

        This is similar to calling ``map`` and ``unwrap_or_else`` in
        succession.

        >>> Maybe(1).map_or_else(lambda n: [n], list)
        [1]
        >>> Maybe(None).map_or_else(lambda n: [n], list)
        []

        """
        return map_or_else(self.val, fn, factory)

    def unwrap(self) -> T:
        """
        Assert that the value is not ``None``.

        >>> Maybe(1).unwrap()
        1
        >>> Maybe(None).unwrap()
        Traceback (most recent call last):
            ...
        ValueError: Trying to unwrap `None` value.

        """
        return unwrap(self.val)

    def unwrap_or(self, default: T) -> T:
        """
        Get ``val`` if it is not ``None``, or ``default`` otherwise.

        >>> Maybe(1).unwrap_or(2)
        1
        >>> Maybe(None).unwrap_or(2)
        2

        """
        return unwrap_or(self.val, default)

    def unwrap_or_else(self, fn: Callable[[], T]) -> T:
        """
        Get ``val`` if it is not ``None``, or invoke ``factory`` to get a
        fallback.

        >>> Maybe([1, 2, 3]).unwrap_or_else(list)
        [1, 2, 3]
        >>> Maybe(None).unwrap_or_else(list)
        []

        """
        return unwrap_or_else(self.val, fn)

    def and_(self, other: Optional[U]) -> Optional[U]:
        """
        Like ``a and b`` in Python, except only considering ``None``.

        This is implement identical to ``unwrap_or``, but different with
        respect to types.

        >>> Maybe(1).and_(2)
        2
        >>> Maybe(1).and_(None)
        >>> Maybe(None).and_(2)

        """
        return and_(self.val, other)

    def __and__(self, other: Optional[U]) -> Optional[U]:
        """
        Like ``a and b`` in Python, except only considering ``None``.

        This is implement identical to ``unwrap_or``, but different with
        respect to types.

        >>> Maybe(1) & 2
        2
        >>> Maybe(1) & None
        >>> Maybe(None) & 2

        """
        return and_(self.val, other)

    def and_then(self, fn: Callable[[T], Optional[U]]) -> Optional[U]:
        """
        Apply ``fn`` to ``val`` if it is not ``None`` and return the result.

        In contrast to ``map``, ``fn`` always returns an ``Optional``, which is
        consequently flattened.

        >>> Maybe([42]).and_then(lambda xs: xs[0] if xs else None)
        42
        >>> Maybe([]).and_then(lambda xs: xs[0] if xs else None)
        >>> Maybe(None).and_then(lambda xs: xs[0] if xs else None)

        """
        return and_then(self.val, fn)

    def or_(self, default: Optional[T]) -> Optional[T]:
        """
        Like ``a or b`` in Python, except only considering ``None``.

        >>> Maybe(1).or_(2)
        1
        >>> Maybe(1).or_(None)
        1
        >>> Maybe(None).or_(2)
        2

        """
        return or_(self.val, default)

    def __or__(self, default: Optional[T]) -> Optional[T]:
        """
        Like ``a or b`` in Python, except only considering ``None``.

        >>> Maybe(1) | 2
        1
        >>> Maybe(1) | None
        1
        >>> Maybe(None) | 2
        2

        """
        return or_(self.val, default)

    def or_else(self, factory: Callable[[], Optional[T]]) -> Optional[T]:
        """
        Like `unwrap_or_else`, except that it returns an optional value.

        >>> Maybe([42]).or_else(list)
        [42]
        >>> Maybe(None).or_else(list)
        []

        """
        return or_else(self.val, factory)

    def contains(self, other: U) -> bool:
        """
        Check if ``val`` equals ``other``, always return ``False`` if ``val``
        is ``None``.

        >>> Maybe(1).contains(1)
        True
        >>> Maybe(1).contains(2)
        False
        >>> Maybe(None).contains(3)
        False

        """
        return contains(self.val, other)

    def filter(self, pred: Callable[[T], bool]) -> Optional[T]:
        """
        Return ``None`` if ``val`` is ``None`` or if ``pred(val)`` does not
        return ``True``, otherwise return ``val``.

        >>> is_even = lambda n: n % 2 == 0
        >>> Maybe(1).filter(is_even)
        >>> Maybe(2).filter(is_even)
        2
        >>> Maybe(None).filter(is_even)

        """
        return filter(self.val, pred)

    def xor(self, other: Optional[T]) -> Optional[T]:
        """
        Return either ``val`` or ``other`` if the other is ``None``. Also
        return ``None`` if both are not ``None``.

        >>> Maybe(1).xor(None)
        1
        >>> Maybe(None).xor(2)
        2
        >>> Maybe(1).xor(2)
        >>> Maybe(None).xor(None)

        """
        return xor(self.val, other)

    def __xor__(self, other: Optional[T]) -> Optional[T]:
        """
        Return either ``val`` or ``other`` if the other is ``None``. Also
        return ``None`` if both are not ``None``.

        >>> Maybe(1) ^ None
        1
        >>> Maybe(None) ^ 2
        2
        >>> Maybe(1) ^ 2
        >>> Maybe(None) ^ None

        """
        return xor(self.val, other)

    def flatten(self: "Maybe[Optional[T]]") -> Optional[T]:
        """Flatten nested optional value.

        Note: This just returns the value, but changes the type from
        ``Optional[Optional[T]]`` to ``Optional[T].``
        """
        return flatten(self.val)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Tuple

import torch
from torch import nn

from gluonts.core.component import validated
from gluonts.model import Input, InputSpec
from gluonts.torch.distributions import StudentTOutput
from gluonts.torch.scaler import StdScaler, MeanScaler, NOPScaler
from gluonts.torch.model.simple_feedforward import make_linear_layer


class MovingAvg(nn.Module):
    """
    Moving average block to highlight the trend of time series
    """

    def __init__(self, kernel_size, stride):
        super().__init__()
        self.kernel_size = kernel_size
        self.avg = nn.AvgPool1d(
            kernel_size=kernel_size, stride=stride, padding=0
        )

    def forward(self, x):
        # padding on the both ends of time series
        front = x[:, 0:1, ...].repeat(1, (self.kernel_size - 1) // 2, 1)
        end = x[:, -1:, ...].repeat(1, (self.kernel_size - 1) // 2, 1)
        x = torch.cat([front, x, end], dim=1)
        x = self.avg(x.permute(0, 2, 1))
        x = x.permute(0, 2, 1)
        return x


class SeriesDecomp(nn.Module):
    """
    Series decomposition block
    """

    def __init__(self, kernel_size):
        super().__init__()
        self.moving_avg = MovingAvg(kernel_size, stride=1)

    def forward(self, x):
        moving_mean = self.moving_avg(x)
        res = x - moving_mean
        return res, moving_mean


class DLinearModel(nn.Module):
    """
    Module implementing a feed-forward model form the paper
    https://arxiv.org/pdf/2205.13504.pdf extended for probabilistic forecasting.

    Parameters
    ----------
    prediction_length
        Number of time points to predict.
    context_length
        Number of time steps prior to prediction time that the model.
    hidden_dimension
        Size of last hidden layers in the feed-forward network.
    distr_output
        Distribution to use to evaluate observations and sample predictions.
    """

    @validated()
    def __init__(
        self,
        prediction_length: int,
        context_length: int,
        hidden_dimension: int,
        distr_output=StudentTOutput(),
        kernel_size: int = 25,
        scaling: str = "mean",
    ) -> None:
        super().__init__()

        assert prediction_length > 0
        assert context_length > 0

        self.prediction_length = prediction_length
        self.context_length = context_length
        self.hidden_dimension = hidden_dimension
        self.decomposition = SeriesDecomp(kernel_size)

        self.distr_output = distr_output
        if scaling == "mean":
            self.scaler = MeanScaler(keepdim=True)
        elif scaling == "std":
            self.scaler = StdScaler(keepdim=True)
        else:
            self.scaler = NOPScaler(keepdim=True)

        self.kernel_size = kernel_size

        self.linear_seasonal = make_linear_layer(
            context_length, prediction_length * hidden_dimension
        )
        self.linear_trend = make_linear_layer(
            context_length, prediction_length * hidden_dimension
        )

        self.args_proj = self.distr_output.get_args_proj(hidden_dimension)

    def describe_inputs(self, batch_size=1) -> InputSpec:
        return InputSpec(
            {
                "past_target": Input(
                    shape=(batch_size, self.context_length), dtype=torch.float
                ),
                "past_observed_values": Input(
                    shape=(batch_size, self.context_length), dtype=torch.float
                ),
            },
            torch.zeros,
        )

    def forward(
        self,
        past_target: torch.Tensor,
        past_observed_values: torch.Tensor,
    ) -> Tuple[Tuple[torch.Tensor, ...], torch.Tensor, torch.Tensor]:
        # scale the input
        past_target_scaled, loc, scale = self.scaler(
            past_target, past_observed_values
        )
        res, trend = self.decomposition(past_target_scaled.unsqueeze(-1))
        seasonal_output = self.linear_seasonal(res.squeeze(-1))
        trend_output = self.linear_trend(trend.squeeze(-1))
        nn_out = seasonal_output + trend_output

        distr_args = self.args_proj(
            nn_out.reshape(-1, self.prediction_length, self.hidden_dimension)
        )
        return distr_args, loc, scale



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional
from typing_extensions import Protocol, runtime_checkable

import gluonts
from gluonts.core.component import from_hyperparameters
from gluonts.dataset.common import Dataset
from gluonts.model.predictor import Predictor


class Estimator:
    """
    An abstract class representing a trainable model.

    The underlying model is trained by calling the `train` method with a
    training `Dataset`, producing a `Predictor` object.
    """

    __version__: str = gluonts.__version__

    prediction_length: int
    lead_time: int

    def __init__(self, lead_time: int = 0, **kwargs) -> None:
        # TODO validation of prediction_length could also
        # TODO be bubbled-up here from subclasses classes
        assert lead_time >= 0, "The value of `lead_time` should be >= 0"

        self.lead_time = lead_time

    def train(
        self, training_data: Dataset, validation_data: Optional[Dataset] = None
    ) -> Predictor:
        """
        Train the estimator on the given data.

        Parameters
        ----------
        training_data
            Dataset to train the model on.
        validation_data
            Dataset to validate the model on during training.

        Returns
        -------
        Predictor
            The predictor containing the trained model.
        """
        raise NotImplementedError

    @classmethod
    def from_hyperparameters(cls, **hyperparameters):
        return from_hyperparameters(cls, **hyperparameters)

    @classmethod
    def derive_auto_fields(cls, train_iter):
        return {}

    @classmethod
    def from_inputs(cls, train_iter, **params):
        # auto_params usually include `use_feat_dynamic_real`,
        # `use_feat_static_cat` and `cardinality`
        auto_params = cls.derive_auto_fields(train_iter)
        # user specified 'params' will take precedence:
        params = {**auto_params, **params}
        return cls.from_hyperparameters(**params)


@runtime_checkable
class IncrementallyTrainable(Protocol):
    def train_from(
        self,
        predictor: Predictor,
        training_data: Dataset,
        validation_data: Optional[Dataset] = None,
    ) -> Predictor:
        """
        Experimental: this feature may change in future versions.
        Train the estimator, starting from a previously trained
        predictor, on the given data.

        Parameters
        ----------
        predictor
            A previously trained model, from which to initialize the estimator
            training.
        training_data
            Dataset to train the model on.
        validation_data
            Dataset to validate the model on during training.

        Returns
        -------
        Predictor
            The predictor containing the trained model.
        """
        raise NotImplementedError



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

import numpy as np

import gluonts.zebras as zb


schema = zb.Schema(
    {
        "target": zb.Field(ndim=1, tdim=0, past_only=True),
        "time_feat": zb.Field(ndim=2, tdim=-1, preprocess=np.atleast_2d),
        "static_feat": zb.Field(ndim=1, preprocess=np.atleast_1d),
    }
)


def test_schema_timeframe():
    xs = [1, 2, 3, 4, 5]
    row = {"target": xs, "time_feat": xs, "static_feat": 1}
    schema.load_timeframe(row)

    row = {"target": xs, "time_feat": [xs, xs], "static_feat": [1]}
    schema.load_timeframe(row)


def test_schema_timeframe_missing():
    xs = [1, 2, 3, 4, 5]
    row = {"target": xs, "time_feat": xs}

    with pytest.raises(Exception):
        schema.load_timeframe({})

    with pytest.raises(Exception):
        schema.load_timeframe(row)


def test_schema_splitframe():
    short = [1, 2, 3, 4, 5]
    long = [1, 2, 3, 4, 5, 6, 7, 8]
    row = {"target": short, "time_feat": long, "static_feat": 1}
    schema.load_splitframe(row)

    row = {"target": short, "time_feat": [long, long], "static_feat": [1]}
    schema.load_splitframe(row)


def test_schema_splitframe_error():
    xs = [1, 2, 3, 4, 5]

    with pytest.raises(Exception):
        schema.load_splitframe({})

    with pytest.raises(Exception):
        row = {"target": xs, "time_feat": xs}
        schema.load_timeframe(row)



from m2r2 import MdInclude

# This is re-exposing `mdinclunde` from m2r2, without registering `.md`, which
# we want to have handled by myst-parser.


def setup(app):
    app.add_directive("mdinclude", MdInclude)
    app.add_config_value("no_underscore_emphasis", False, "env")
    app.add_config_value("m2r_parse_relative_links", False, "env")
    app.add_config_value("m2r_anonymous_references", False, "env")
    app.add_config_value("m2r_disable_inline_math", False, "env")
    app.add_config_value(
        "m2r_use_mermaid",
        "sphinxcontrib.mermaid" in app.config.extensions,
        "env",
    )

    return {
        "version": "0.0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple

import mxnet as mx

from gluonts.core.component import validated
from gluonts.mx import Tensor
from gluonts.mx.block.scaler import MeanScaler, NOPScaler
from gluonts.mx.distribution import Distribution, DistributionOutput
from gluonts.mx.util import assert_shape, weighted_average


def make_rnn_cell(
    num_cells: int,
    num_layers: int,
    cell_type: str,
    residual: bool,
    dropout_rate: float,
) -> mx.gluon.HybridBlock:
    RnnCell = {"lstm": mx.gluon.rnn.LSTMCell, "gru": mx.gluon.rnn.GRUCell}[
        cell_type
    ]
    rnn = mx.gluon.rnn.HybridSequentialRNNCell()
    for k in range(num_layers):
        cell = RnnCell(hidden_size=num_cells)
        if residual:
            cell = mx.gluon.rnn.ResidualCell(cell) if k > 0 else cell
        cell = (
            mx.gluon.rnn.ZoneoutCell(cell, zoneout_states=dropout_rate)
            if dropout_rate > 0.0
            else cell
        )
        rnn.add(cell)
    return rnn


class DeepVARNetwork(mx.gluon.HybridBlock):
    @validated()
    def __init__(
        self,
        num_layers: int,
        num_cells: int,
        cell_type: str,
        history_length: int,
        context_length: int,
        prediction_length: int,
        distr_output: DistributionOutput,
        dropout_rate: float,
        lags_seq: List[int],
        target_dim: int,
        cardinality: List[int] = [1],
        embedding_dimension: int = 1,
        scaling: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.num_layers = num_layers
        self.num_cells = num_cells
        self.cell_type = cell_type
        self.history_length = history_length
        self.context_length = context_length
        self.prediction_length = prediction_length
        self.dropout_rate = dropout_rate
        self.cardinality = cardinality
        self.embedding_dimension = embedding_dimension
        self.num_cat = len(cardinality)
        self.target_dim = target_dim
        self.scaling = scaling
        self.target_dim_sample = target_dim

        assert len(set(lags_seq)) == len(
            lags_seq
        ), "no duplicated lags allowed!"
        lags_seq.sort()

        self.lags_seq = lags_seq

        self.distr_output = distr_output

        self.target_dim = target_dim

        with self.name_scope():
            self.proj_dist_args = distr_output.get_args_proj()

            residual = True

            self.rnn = make_rnn_cell(
                cell_type=cell_type,
                num_cells=num_cells,
                num_layers=num_layers,
                residual=residual,
                dropout_rate=dropout_rate,
            )

            self.embed_dim = 1
            self.embed = mx.gluon.nn.Embedding(
                input_dim=self.target_dim, output_dim=self.embed_dim
            )

            if scaling:
                self.scaler = MeanScaler(keepdims=True)
            else:
                self.scaler = NOPScaler(keepdims=True)

    @staticmethod
    def get_lagged_subsequences(
        F,
        sequence: Tensor,
        sequence_length: int,
        indices: List[int],
        subsequences_length: int = 1,
    ) -> Tensor:
        """
        Returns lagged subsequences of a given sequence.
        Parameters
        ----------
        sequence
            the sequence from which lagged subsequences should be extracted.
            Shape: (N, T, C).
        sequence_length
            length of sequence in the T (time) dimension (axis = 1).
        indices
            list of lag indices to be used.
        subsequences_length
            length of the subsequences to be extracted.
        Returns
        --------
        lagged : Tensor
            a tensor of shape (N, S, C, I),
            where S = subsequences_length and I = len(indices),
            containing lagged subsequences.
            Specifically, lagged[i, :, j, k] = sequence[i, -indices[k]-S+j, :].
        """
        # we must have: history_length + begin_index >= 0
        # that is: history_length - lag_index - sequence_length >= 0
        # hence the following assert
        assert max(indices) + subsequences_length <= sequence_length, (
            "lags cannot go further than history length, found lag "
            f"{max(indices)} while history length is only {sequence_length}"
        )
        assert all(lag_index >= 0 for lag_index in indices)

        lagged_values = []
        for lag_index in indices:
            begin_index = -lag_index - subsequences_length
            end_index = -lag_index if lag_index > 0 else None
            lagged_values.append(
                F.slice_axis(
                    sequence, axis=1, begin=begin_index, end=end_index
                ).expand_dims(axis=1)
            )
        return F.concat(
            *lagged_values, num_args=len(indices), dim=1
        ).transpose(axes=(0, 2, 3, 1))

    def unroll(
        self,
        F,
        lags: Tensor,
        scale: Tensor,
        time_feat: Tensor,
        target_dimension_indicator: Tensor,
        unroll_length: int,
        begin_state: Optional[List[Tensor]],
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        """
        Prepares the input to the RNN and unrolls it the given number of time
        steps.

        Parameters
        ----------
        F
        lags
            Input lags (batch_size, sub_seq_len, target_dim, num_lags)
        scale
            Mean scale (batch_size, 1, target_dim)
        time_feat
            Additional time features
        target_dimension_indicator
            Indices of the target dimension (batch_size, target_dim)
        unroll_length
            length to unroll
        begin_state
            State to start the unrolling of the RNN

        Returns
        -------
        outputs
            RNN outputs (batch_size, seq_len, num_cells)
        states
            RNN states. Nested list with (batch_size, num_cells) tensors with
        dimensions target_dim x num_layers x (batch_size, num_cells)
        lags_scaled
            Scaled lags(batch_size, sub_seq_len, target_dim, num_lags)
        inputs
            inputs to the RNN
        """
        # (batch_size, sub_seq_len, target_dim, num_lags)
        lags_scaled = F.broadcast_div(lags, scale.expand_dims(axis=-1))

        assert_shape(
            lags_scaled,
            (-1, unroll_length, self.target_dim, len(self.lags_seq)),
        )

        input_lags = F.reshape(
            data=lags_scaled,
            shape=(-1, unroll_length, len(self.lags_seq) * self.target_dim),
        )

        # (batch_size, target_dim, embed_dim)
        index_embeddings = self.embed(target_dimension_indicator)
        assert_shape(index_embeddings, (-1, self.target_dim, self.embed_dim))

        # (batch_size, seq_len, target_dim * embed_dim)
        repeated_index_embeddings = (
            index_embeddings.expand_dims(axis=1)
            .repeat(axis=1, repeats=unroll_length)
            .reshape((-1, unroll_length, self.target_dim * self.embed_dim))
        )

        # (batch_size, sub_seq_len, input_dim)
        inputs = F.concat(
            input_lags, repeated_index_embeddings, time_feat, dim=-1
        )

        # unroll encoder
        outputs, state = self.rnn.unroll(
            inputs=inputs,
            length=unroll_length,
            layout="NTC",
            merge_outputs=True,
            begin_state=begin_state,
        )

        assert_shape(outputs, (-1, unroll_length, self.num_cells))
        for s in state:
            assert_shape(s, (-1, self.num_cells))

        assert_shape(
            lags_scaled,
            (-1, unroll_length, self.target_dim, len(self.lags_seq)),
        )

        return outputs, state, lags_scaled, inputs

    def unroll_encoder(
        self,
        F,
        past_time_feat: Tensor,
        past_target_cdf: Tensor,
        past_observed_values: Tensor,
        past_is_pad: Tensor,
        future_time_feat: Optional[Tensor],
        future_target_cdf: Optional[Tensor],
        target_dimension_indicator: Tensor,
    ) -> Tuple[Tensor, List[Tensor], Tensor, Tensor, Tensor]:
        """
        Unrolls the RNN encoder over past and, if present, future data. Returns
        outputs and state of the encoder, plus the scale of past_target_cdf and
        a vector of static features that was constructed and fed as input to
        the encoder. All tensor arguments should have NTC layout.

        Parameters
        ----------
        F
        past_time_feat
            Past time features (batch_size, history_length, num_features)
        past_target_cdf
            Past marginal CDF transformed target values (batch_size,
            history_length, target_dim)
        past_observed_values
            Indicator whether or not the values were observed (batch_size,
            history_length, target_dim)
        past_is_pad
            Indicator whether the past target values have been padded
            (batch_size, history_length)
        future_time_feat
            Future time features (batch_size, prediction_length, num_features)
        future_target_cdf
            Future marginal CDF transformed target values (batch_size,
            prediction_length, target_dim)
        target_dimension_indicator
            Dimensionality of the time series (batch_size, target_dim)

        Returns
        -------
        outputs
            RNN outputs (batch_size, seq_len, num_cells)
        states
            RNN states. Nested list with (batch_size, num_cells) tensors with
        dimensions target_dim x num_layers x (batch_size, num_cells)
        scale
            Mean scales for the time series (batch_size, 1, target_dim)
        lags_scaled
            Scaled lags(batch_size, sub_seq_len, target_dim, num_lags)
        inputs
            inputs to the RNN
        """

        past_observed_values = F.broadcast_minimum(
            past_observed_values, 1 - past_is_pad.expand_dims(axis=-1)
        )

        if future_time_feat is None or future_target_cdf is None:
            time_feat = past_time_feat.slice_axis(
                axis=1, begin=-self.context_length, end=None
            )
            sequence = past_target_cdf
            sequence_length = self.history_length
            subsequences_length = self.context_length
        else:
            time_feat = F.concat(
                past_time_feat.slice_axis(
                    axis=1, begin=-self.context_length, end=None
                ),
                future_time_feat,
                dim=1,
            )
            sequence = F.concat(past_target_cdf, future_target_cdf, dim=1)
            sequence_length = self.history_length + self.prediction_length
            subsequences_length = self.context_length + self.prediction_length

        # (batch_size, sub_seq_len, target_dim, num_lags)
        lags = self.get_lagged_subsequences(
            F=F,
            sequence=sequence,
            sequence_length=sequence_length,
            indices=self.lags_seq,
            subsequences_length=subsequences_length,
        )

        # scale is computed on the context length last units of the past target
        # scale shape is (batch_size, 1, target_dim)
        _, scale = self.scaler(
            past_target_cdf.slice_axis(
                axis=1, begin=-self.context_length, end=None
            ),
            past_observed_values.slice_axis(
                axis=1, begin=-self.context_length, end=None
            ),
        )

        outputs, states, lags_scaled, inputs = self.unroll(
            F=F,
            lags=lags,
            scale=scale,
            time_feat=time_feat,
            target_dimension_indicator=target_dimension_indicator,
            unroll_length=subsequences_length,
            begin_state=None,
        )

        return outputs, states, scale, lags_scaled, inputs

    def distr(
        self,
        rnn_outputs: Tensor,
        time_features: Tensor,
        scale: Tensor,
        lags_scaled: Tensor,
        target_dimension_indicator: Tensor,
        seq_len: int,
    ):
        """
        Returns the distribution of DeepVAR with respect to the RNN outputs.

        Parameters
        ----------
        rnn_outputs
            Outputs of the unrolled RNN (batch_size, seq_len, num_cells)
        time_features
            Dynamic time features (batch_size, seq_len, num_features)
        scale
            Mean scale for each time series (batch_size, 1, target_dim)
        lags_scaled
            Scaled lags used for RNN input
            (batch_size, seq_len, target_dim, num_lags)
        target_dimension_indicator
            Indices of the target dimension (batch_size, target_dim)
        seq_len
            Length of the sequences

        Returns
        -------
        distr
            Distribution instance
        distr_args
            Distribution arguments
        """
        distr_args = self.proj_dist_args(rnn_outputs)

        # compute likelihood of target given the predicted parameters
        distr = self.distr_output.distribution(distr_args, scale=scale)

        return distr, distr_args

    def loss(self, F, target: Tensor, distr: Distribution) -> Tensor:
        """
        Returns negative log-likelihood of `target` under `distr`.

        Parameters
        ----------
        F
        target
            Tensor with shape (batch_size, seq_len, target_dim)
        distr
            Distribution instances

        Returns
        -------
        Loss
            Tensor with shape (batch_size, seq_length, 1)
        """
        # we sum the last axis to have the same shape for all likelihoods
        # (batch_size, subseq_length, 1)
        return -distr.log_prob(target).expand_dims(axis=-1)

    def train_hybrid_forward(
        self,
        F,
        target_dimension_indicator: Tensor,
        past_time_feat: Tensor,
        past_target_cdf: Tensor,
        past_observed_values: Tensor,
        past_is_pad: Tensor,
        future_time_feat: Tensor,
        future_target_cdf: Tensor,
        future_observed_values: Tensor,
    ) -> Tuple[Tensor, ...]:
        """
        Computes the loss for training DeepVAR, all inputs tensors representing
        time series have NTC layout.

        Parameters
        ----------
        F
        target_dimension_indicator
            Indices of the target dimension (batch_size, target_dim)
        past_time_feat
            Dynamic features of past time series (batch_size, history_length,
            num_features)
        past_target_cdf
            Past marginal CDF transformed target values (batch_size,
            history_length, target_dim)
        past_observed_values
            Indicator whether or not the values were observed (batch_size,
            history_length, target_dim)
        past_is_pad
            Indicator whether the past target values have been padded
            (batch_size, history_length)
        future_time_feat
            Future time features (batch_size, prediction_length, num_features)
        future_target_cdf
            Future marginal CDF transformed target values (batch_size,
            prediction_length, target_dim)
        future_observed_values
            Indicator whether or not the future values were observed
            (batch_size, prediction_length, target_dim)

        Returns
        -------
        distr
            Loss with shape (batch_size, 1)
        likelihoods
            Likelihoods for each time step
            (batch_size, context + prediction_length, 1)
        distr_args
            Distribution arguments (context + prediction_length,
            number_of_arguments)
        """

        seq_len = self.context_length + self.prediction_length

        # unroll the decoder in "training mode", i.e. by providing future data
        # as well
        rnn_outputs, _, scale, lags_scaled, inputs = self.unroll_encoder(
            F=F,
            past_time_feat=past_time_feat,
            past_target_cdf=past_target_cdf,
            past_observed_values=past_observed_values,
            past_is_pad=past_is_pad,
            future_time_feat=future_time_feat,
            future_target_cdf=future_target_cdf,
            target_dimension_indicator=target_dimension_indicator,
        )

        # put together target sequence
        # (batch_size, seq_len, target_dim)
        target = F.concat(
            past_target_cdf.slice_axis(
                axis=1, begin=-self.context_length, end=None
            ),
            future_target_cdf,
            dim=1,
        )

        # assert_shape(target, (-1, seq_len, self.target_dim))

        distr, distr_args = self.distr(
            time_features=inputs,
            rnn_outputs=rnn_outputs,
            scale=scale,
            lags_scaled=lags_scaled,
            target_dimension_indicator=target_dimension_indicator,
            seq_len=self.context_length + self.prediction_length,
        )

        loss = self.loss(F, target=target, distr=distr)
        assert_shape(loss, (-1, seq_len, 1))

        past_observed_values = F.broadcast_minimum(
            past_observed_values, 1 - past_is_pad.expand_dims(axis=-1)
        )

        # (batch_size, subseq_length, target_dim)
        observed_values = F.concat(
            past_observed_values.slice_axis(
                axis=1, begin=-self.context_length, end=None
            ),
            future_observed_values,
            dim=1,
        )

        # mask the loss at one time step if one or more observations is missing
        # in the target dimensions (batch_size, subseq_length, 1)
        loss_weights = observed_values.min(axis=-1, keepdims=True)

        assert_shape(loss_weights, (-1, seq_len, 1))

        weighted_loss = weighted_average(
            F=F, x=loss, weights=loss_weights, axis=1
        )

        assert_shape(weighted_loss, (-1, -1, 1))

        self.distribution = distr

        return (weighted_loss, loss) + distr_args

    def sampling_decoder(
        self,
        F,
        past_target_cdf: Tensor,
        target_dimension_indicator: Tensor,
        time_feat: Tensor,
        scale: Tensor,
        begin_states: List[Tensor],
    ) -> Tensor:
        """
        Computes sample paths by unrolling the RNN starting with a initial
        input and state.

        Parameters
        ----------
        past_target_cdf
            Past marginal CDF transformed target values (batch_size,
            history_length, target_dim)
        target_dimension_indicator
            Indices of the target dimension (batch_size, target_dim)
        time_feat
            Dynamic features of future time series (batch_size, history_length,
            num_features)
        scale
            Mean scale for each time series (batch_size, 1, target_dim)
        begin_states
            List of initial states for the RNN layers (batch_size, num_cells)
        Returns
        --------
        sample_paths : Tensor
            A tensor containing sampled paths. Shape: (1, num_sample_paths,
            prediction_length, target_dim).
        """

        def repeat(tensor):
            return tensor.repeat(repeats=self.num_parallel_samples, axis=0)

        # blows-up the dimension of each tensor to
        # batch_size * self.num_sample_paths for increasing parallelism
        repeated_past_target_cdf = repeat(past_target_cdf)
        repeated_time_feat = repeat(time_feat)
        repeated_scale = repeat(scale)
        repeated_target_dimension_indicator = repeat(
            target_dimension_indicator
        )

        # slight difference for GPVAR and DeepVAR, in GPVAR, its a list
        repeated_states = self.make_states(begin_states)

        future_samples = []

        # for each future time-units we draw new samples for this time-unit
        # and update the state
        for k in range(self.prediction_length):
            lags = self.get_lagged_subsequences(
                F=F,
                sequence=repeated_past_target_cdf,
                sequence_length=self.history_length + k,
                indices=self.shifted_lags,
                subsequences_length=1,
            )

            rnn_outputs, repeated_states, lags_scaled, inputs = self.unroll(
                F=F,
                begin_state=repeated_states,
                lags=lags,
                scale=repeated_scale,
                time_feat=repeated_time_feat.slice_axis(
                    axis=1, begin=k, end=k + 1
                ),
                target_dimension_indicator=repeated_target_dimension_indicator,
                unroll_length=1,
            )

            distr, distr_args = self.distr(
                time_features=inputs,
                rnn_outputs=rnn_outputs,
                scale=repeated_scale,
                target_dimension_indicator=repeated_target_dimension_indicator,
                lags_scaled=lags_scaled,
                seq_len=1,
            )

            # (batch_size, 1, target_dim)
            new_samples = distr.sample()
            new_samples = self.post_process_samples(new_samples)

            # (batch_size, seq_len, target_dim)
            future_samples.append(new_samples)
            repeated_past_target_cdf = F.concat(
                repeated_past_target_cdf, new_samples, dim=1
            )

        # (batch_size * num_samples, prediction_length, target_dim)
        samples = F.concat(*future_samples, dim=1)

        # (batch_size, num_samples, prediction_length, target_dim)
        return samples.reshape(
            shape=(
                -1,
                self.num_parallel_samples,
                self.prediction_length,
                self.target_dim,
            )
        )

    def make_states(self, begin_states: List[Tensor]) -> List[Tensor]:
        """
        Repeat states to match the the shape induced by the number of sample
        paths.

        Parameters
        ----------
        begin_states
            List of initial states for the RNN layers (batch_size, num_cells)

        Returns
        -------
            List of initial states
        """

        def repeat(tensor):
            return tensor.repeat(repeats=self.num_parallel_samples, axis=0)

        return [repeat(s) for s in begin_states]

    def predict_hybrid_forward(
        self,
        F,
        target_dimension_indicator: Tensor,
        past_time_feat: Tensor,
        past_target_cdf: Tensor,
        past_observed_values: Tensor,
        past_is_pad: Tensor,
        future_time_feat: Tensor,
    ) -> Tensor:
        """
        Predicts samples given the trained DeepVAR model.
        All tensors should have NTC layout.
        Parameters
        ----------
        F
        target_dimension_indicator
            Indices of the target dimension (batch_size, target_dim)
        past_time_feat
            Dynamic features of past time series (batch_size, history_length,
            num_features)
        past_target_cdf
            Past marginal CDF transformed target values (batch_size,
            history_length, target_dim)
        past_observed_values
            Indicator whether or not the values were observed (batch_size,
            history_length, target_dim)
        past_is_pad
            Indicator whether the past target values have been padded
            (batch_size, history_length)
        future_time_feat
            Future time features (batch_size, prediction_length, num_features)

        Returns
        -------
        sample_paths : Tensor
            A tensor containing sampled paths (1, num_sample_paths,
            prediction_length, target_dim).

        """

        # mark padded data as unobserved
        # (batch_size, target_dim, seq_len)
        past_observed_values = F.broadcast_minimum(
            past_observed_values, 1 - past_is_pad.expand_dims(axis=-1)
        )

        # unroll the decoder in "prediction mode", i.e. with past data only
        _, state, scale, _, inputs = self.unroll_encoder(
            F=F,
            past_time_feat=past_time_feat,
            past_target_cdf=past_target_cdf,
            past_observed_values=past_observed_values,
            past_is_pad=past_is_pad,
            future_time_feat=None,
            future_target_cdf=None,
            target_dimension_indicator=target_dimension_indicator,
        )

        return self.sampling_decoder(
            F=F,
            past_target_cdf=past_target_cdf,
            target_dimension_indicator=target_dimension_indicator,
            time_feat=future_time_feat,
            scale=scale,
            begin_states=state,
        )

    def post_process_samples(self, samples: Tensor) -> Tensor:
        """
        Method to enforce domain-specific constraints on the generated samples.

        For example, see `DeepVARHierarchicalNetwork`, which adjusts the
        samples so that they satisfy the given aggregation constraints.

        For `DeepVAR` this is simply the Identity map.

        Parameters
        ----------
        samples
            Tensor of shape (num_parallel_samples*batch_size, 1, target_dim)

        Returns
        -------
            Tensor of samples with the same shape.
        """
        return samples


class DeepVARTrainingNetwork(DeepVARNetwork):
    def hybrid_forward(
        self,
        F,
        target_dimension_indicator: Tensor,
        past_time_feat: Tensor,
        past_target_cdf: Tensor,
        past_observed_values: Tensor,
        past_is_pad: Tensor,
        future_time_feat: Tensor,
        future_target_cdf: Tensor,
        future_observed_values: Tensor,
    ) -> Tuple[Tensor, ...]:
        """
        Computes the loss for training DeepVAR, all inputs tensors representing
        time series have NTC layout.

        Parameters
        ----------
        F
        target_dimension_indicator
            Indices of the target dimension (batch_size, target_dim)
        past_time_feat
            Dynamic features of past time series (batch_size, history_length,
            num_features)
        past_target_cdf
            Past marginal CDF transformed target values (batch_size,
            history_length, target_dim)
        past_observed_values
            Indicator whether or not the values were observed (batch_size,
            history_length, target_dim)
        past_is_pad
            Indicator whether the past target values have been padded
            (batch_size, history_length)
        future_time_feat
            Future time features (batch_size, prediction_length, num_features)
        future_target_cdf
            Future marginal CDF transformed target values (batch_size,
            prediction_length, target_dim)
        future_observed_values
            Indicator whether or not the future values were observed
            (batch_size, prediction_length, target_dim)

        Returns
        -------
        distr
            Loss with shape (batch_size, 1)
        likelihoods
            Likelihoods for each time step
            (batch_size, context + prediction_length, 1)
        distr_args
            Distribution arguments (context + prediction_length,
            number_of_arguments)
        """
        return self.train_hybrid_forward(
            F,
            target_dimension_indicator,
            past_time_feat,
            past_target_cdf,
            past_observed_values,
            past_is_pad,
            future_time_feat,
            future_target_cdf,
            future_observed_values,
        )


class DeepVARPredictionNetwork(DeepVARNetwork):
    @validated()
    def __init__(self, num_parallel_samples: int, **kwargs) -> None:
        super().__init__(**kwargs)
        self.num_parallel_samples = num_parallel_samples

        # for decoding the lags are shifted by one,
        # at the first time-step of the decoder a lag of one corresponds to
        # the last target value
        self.shifted_lags = [l - 1 for l in self.lags_seq]

    def hybrid_forward(
        self,
        F,
        target_dimension_indicator: Tensor,
        past_time_feat: Tensor,
        past_target_cdf: Tensor,
        past_observed_values: Tensor,
        past_is_pad: Tensor,
        future_time_feat: Tensor,
    ) -> Tensor:
        """
        Predicts samples given the trained DeepVAR model.
        All tensors should have NTC layout.
        Parameters
        ----------
        F
        target_dimension_indicator
            Indices of the target dimension (batch_size, target_dim)
        past_time_feat
            Dynamic features of past time series (batch_size, history_length,
            num_features)
        past_target_cdf
            Past marginal CDF transformed target values (batch_size,
            history_length, target_dim)
        past_observed_values
            Indicator whether or not the values were observed (batch_size,
            history_length, target_dim)
        past_is_pad
            Indicator whether the past target values have been padded
            (batch_size, history_length)
        future_time_feat
            Future time features (batch_size, prediction_length, num_features)

        Returns
        -------
        sample_paths : Tensor
            A tensor containing sampled paths (1, num_sample_paths,
            prediction_length, target_dim).

        """
        return self.predict_hybrid_forward(
            F=F,
            target_dimension_indicator=target_dimension_indicator,
            past_time_feat=past_time_feat,
            past_target_cdf=past_target_cdf,
            past_observed_values=past_observed_values,
            past_is_pad=past_is_pad,
            future_time_feat=future_time_feat,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from enum import Enum
from typing import Any, Iterator, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from gluonts.core.component import validated
from gluonts.dataset.common import Dataset
from gluonts.exceptions import GluonTSDataError
from gluonts.model.forecast import SampleForecast
from gluonts.model.predictor import RepresentablePredictor
from gluonts.time_feature import time_features_from_frequency_str

from ._model import NPTS


class KernelType(str, Enum):
    exponential = "exponential"
    uniform = "uniform"


class NPTSPredictor(RepresentablePredictor):
    r"""
    Implementation of Non-Parametric Time Series Forecaster.

    Forecasts of NPTS for time step :math:`T` are one of the previous values
    of the time series (these could be known values or predictions), sampled
    according to the (un-normalized) distribution :math:`q_T(t) > 0`, where
    :math:`0 <= t < T`.

    The distribution :math:`q_T` is expressed in terms of a feature map
    :math:`f(t)` which associates a time step :math:`t` with a
    :math:`D`-dimensional feature map :math:`[f_1(t), ..., f_D(t)]`. More
    details on the feature map can be found below.

    We offer two types of distribution kernels.

    **Exponential Kernel (NPTS Forecaster)**

      The sampling distribution :math:`q_T` for the `exponential` kernel
      can be `weighted` or `unweighted` and is defined as follows.

      .. math::

        q_T(t) =
        \begin{cases}
          \exp( - \sum_{i=1}^D \alpha   \left| f_i(t) - f_i(T) \right| )
            & \text{unweighted}\\
          \exp( - \sum_{i=1}^D \alpha_i \left| f_i(t) - f_i(T) \right| )
            & \text{weighted}
        \end{cases}

      In the above definition :math:`\alpha > 0` and :math:`\alpha_i > 0` are
      user-defined sampling weights.

    **Uniform Kernel (Climatological Forecaster)**

      The sampling distribution :math:`q_T` for the `uniform` kernel can be
      `seasonal` or not. The `seasonal` version is defined as follows.

      .. math::

         q_T(t) =
         \begin{cases}
           1.0
             & \text{if }f(t) = f(T) \\
           0.0
             & \text{otherwise}
         \end{cases}

      The `not seasonal` version is defined as the constant map.

      .. math::

         q_T(t) = 1.0

    **Feature Map**

      The feature map :math:`f` is configurable. The special case
      :math:`f(t) = [t]` results in the so-called `naive NPTS`. For
      non-seasonal models, by default we have :math:`f(t) = [t]` for the NPTS
      Forecaster (i.e., with the `exponential` kernel) and no features for the
      Climatological Forecaster (i.e., the `uniform` kernel).

      For seasonal NPTS and seasonal Climatological, time features determined
      based on the frequency of the time series are added to the default
      feature map.

      The default time features for various frequencies are

      .. math::

         f(t) =
         \begin{cases}
           [\mathit{MINUTE\_OF\_HOUR}(t)] & \text{for minutely frequency}\\
           [\mathit{HOUR\_OF\_DAY}(t)]    & \text{for hourly frequency}\\
           [\mathit{DAY\_OF\_WEEK}(t)]    & \text{for daily frequency}\\
           [\mathit{DAY\_OF\_MONTH}(t)]   & \text{for weekly frequency}\\
           [\mathit{MONTH\_OF\_YEAR}(t)]  & \text{for monthly frequency}
         \end{cases}

      During prediction, one can provide custom features in `feat_dynamic_real`
      (these have to be defined in both the training and the prediction range).
      If the model is seasonal, these custom features are added to the default
      feature map, otherwise they are ignored. If `feat_dynamic_real` is not
      empty, one can disable default time features by setting the flag
      `use_default_time_features` to `False`.

    Parameters
    ----------

    freq
        time frequency string
    prediction_length
        number of time steps to predict
    context_length
        number of time-steps that are considered before making predictions
        (the default value of None corresponds to the case where all time steps
        in the history are considered)
    kernel_type
        the type of kernel to use (either "exponential" or "uniform")
    exp_kernel_weights
        single weight :math:`\alpha` or the weights for the features to use
        in the exponential kernel; currently, we use the single weight version
        and for seasonal NPTS we just rescale :math:`\alpha` by `feature_scale`
        for seasonal features.
    use_seasonal_model
        whether to use seasonal variant
    use_default_time_features
        time features derived based on the frequency of the time series
    num_default_time_features
        this is not exposed; this parameter is for having more control on the
        number of default time features, as the date_feature_set adds too
        many per default.
    feature_scale
        scale for time (seasonal) features in order to sample past seasons
        with higher probability
    """

    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        context_length: Optional[int] = None,
        kernel_type: KernelType = KernelType.exponential,
        exp_kernel_weights: Union[float, List[float]] = 1.0,
        use_seasonal_model: bool = True,
        use_default_time_features: bool = True,
        num_default_time_features: int = 1,
        feature_scale: float = 1000.0,
    ) -> None:
        super().__init__(prediction_length=prediction_length)
        # We limit the context length to some maximum value instead of
        # looking at the whole history which might be too large.
        self.context_length = (
            context_length if context_length is not None else 1100
        )
        self.kernel_type = kernel_type
        self.num_default_time_features = num_default_time_features
        self.use_seasonal_model = use_seasonal_model
        self.use_default_time_features = use_default_time_features
        self.feature_scale = feature_scale
        self.freq = freq

        if not self._is_exp_kernel():
            self.kernel = NPTS.uniform_kernel()
        elif isinstance(exp_kernel_weights, float):
            self.kernel = NPTS.log_distance_kernel(exp_kernel_weights)
        elif isinstance(exp_kernel_weights, list):
            self.kernel = NPTS.log_weighted_distance_kernel(exp_kernel_weights)
        else:
            raise RuntimeError(
                'Unexpected "exp_kernel_weights" type - should be either'
                "a float or a list of floats"
            )

    def _is_exp_kernel(self) -> bool:
        return self.kernel_type == KernelType.exponential

    def predict(
        self, dataset: Dataset, num_samples: int = 100, **kwargs
    ) -> Iterator[SampleForecast]:
        for data in dataset:
            start = data["start"]
            target = np.asarray(data["target"], np.float32)
            index = pd.period_range(
                start=start, freq=self.freq, periods=len(target)
            )
            item_id = data.get("item_id", None)

            # Slice the time series until context_length or history length
            # depending on which ever is minimum
            train_length = min(len(target), self.context_length)
            ts = pd.Series(index=index, data=target)[-train_length:]

            custom_features: Optional[np.ndarray]
            if "feat_dynamic_real" in data.keys():
                custom_features = np.array(
                    [
                        dynamic_feature[
                            -train_length - self.prediction_length :
                        ]
                        for dynamic_feature in data["feat_dynamic_real"]
                    ]
                )
            else:
                custom_features = None

            yield self.predict_time_series(
                ts, num_samples, custom_features, item_id=item_id
            )

    def predict_time_series(
        self,
        ts: pd.Series,
        num_samples: int,
        custom_features: Optional[np.ndarray] = None,
        item_id: Optional[Any] = None,
    ) -> SampleForecast:
        """
        Given a training time series, this method generates `Forecast` object
        containing prediction samples for `prediction_length` time points.

        The predictions are generated via weighted sampling where the weights
        are determined by the `NPTSPredictor` kernel type and feature map.

        Parameters
        ----------
        ts
            training time series object
        custom_features
            custom features (covariates) to use
        num_samples
            number of samples to draw
        item_id
            item_id to identify the time series
        Returns
        -------
        Forecast
          A prediction for the supplied `ts` and `custom_features`.
        """

        if np.all(np.isnan(ts.values[-self.context_length :])):
            raise GluonTSDataError(
                f"The last {self.context_length} positions of the target time "
                "series are all NaN. Please increase the `context_length` "
                "parameter of your NPTS model so the last "
                f"{self.context_length} positions of each target contain at "
                "least one non-NaN value."
            )

        # Get the features for both training and prediction ranges
        train_features, predict_features = self._get_features(
            ts.index, self.prediction_length, custom_features
        )

        # Compute weights for sampling for each time step `t` in the
        # prediction range
        sampling_weights_iterator = NPTS.compute_weights(
            train_features=train_features,
            pred_features=predict_features,
            target_isnan_positions=np.argwhere(np.isnan(ts.values)),
            kernel=self.kernel,
            do_exp=self._is_exp_kernel(),
        )

        # Generate forecasts
        forecast = NPTS.predict(
            targets=ts,
            prediction_length=self.prediction_length,
            sampling_weights_iterator=sampling_weights_iterator,
            num_samples=num_samples,
            item_id=item_id,
        )

        return forecast

    def _get_features(
        self,
        train_index: pd.PeriodIndex,
        prediction_length: int,
        custom_features: Optional[np.ndarray] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Internal method for computing default, (optional) seasonal features for
        the training and prediction ranges given time index for the training
        range and the prediction length.

        Appends `custom_features` if provided.

        Parameters
        ----------
        train_index
            Pandas PeriodIndex
        prediction_length
            prediction length
        custom_features
            shape: (num_custom_features, train_length + pred_length)

        Returns
        -------
        a tuple of (training, prediction) feature tensors
            shape: (num_features, train_length/pred_length)
        """

        train_length = len(train_index)
        full_time_index = pd.period_range(
            train_index.min(),
            periods=train_length + prediction_length,
            freq=train_index.freq,
        )

        # Default feature map for both seasonal and non-seasonal models.
        if self._is_exp_kernel():
            # Default time index features: index of the time point
            # [0, train_length + pred_length - 1]
            features = np.expand_dims(
                np.array(range(len(full_time_index))), axis=0
            )

            # Rescale time index features into the range: [-0.5, 0.5]
            # similar to the seasonal features
            # (see gluonts.time_feature)
            features = features / (train_length + prediction_length - 1) - 0.5
        else:
            # For uniform seasonal model we do not add time index features
            features = np.empty((0, len(full_time_index)))

        # Add more features for seasonal variant
        if self.use_seasonal_model:
            if custom_features is not None:
                total_length = train_length + prediction_length

                assert len(custom_features.shape) == 2, (
                    "Custom features should be 2D-array where the rows "
                    "represent features and columns the time points."
                )

                assert custom_features.shape[1] == total_length, (
                    "For a seasonal model, feat_dynamic_real must be defined "
                    "for both training and prediction ranges. They are only "
                    f"provided for {custom_features.shape[1]} time steps "
                    f"instead of {train_length + prediction_length} steps."
                )

                features = np.vstack(
                    [features, self.feature_scale * custom_features]
                )

            if self.use_default_time_features or custom_features is None:
                # construct seasonal features
                seasonal_features_gen = time_features_from_frequency_str(
                    full_time_index.freqstr
                )
                seasonal_features = [
                    self.feature_scale * gen(full_time_index)
                    for gen in seasonal_features_gen[
                        : self.num_default_time_features
                    ]
                ]

                features = np.vstack([features, *seasonal_features])

        train_features = features[:, :train_length]
        pred_features = features[:, train_length:]

        return train_features, pred_features



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Union, Any, Dict

import numpy as np
import pandas as pd

from statsforecast import StatsForecast
from hierarchicalforecast.core import HierarchicalReconciliation
from hierarchicalforecast.core import _build_fn_name

from gluonts.core.component import validated
from gluonts.dataset import DataEntry
from gluonts.dataset.util import forecast_start
from gluonts.model.predictor import RepresentablePredictor
from gluonts.model.forecast import QuantileForecast
from gluonts.ext.statsforecast import ModelConfig

models_without_fitted_capability = [
    "ADIDA",
    "CrostonClassic",
    "CrostonOptimized",
    "CrostonSBA",
    "IMAPA",
    "SeasWA",
    "TSB",
    "WindowAverage",
]


def get_formatted_S(
    _S: Union[List[List[int]], np.ndarray],
    ts_names: List[str],
) -> pd.DataFrame:
    """
    We format the summation matrix S as a dataframe,
    where the index and columns have the
    corresponding time series names.
    """

    S = np.array(_S)

    return pd.DataFrame(
        S, index=ts_names, columns=ts_names[S.shape[0] - S.shape[1] :]
    )


def format_data_entry(entry: DataEntry, S: pd.DataFrame) -> pd.DataFrame:
    """
    Format data entry as required by hierarchicalforecast.

    ``entry`` is a dictionary with keys: ``"start"``, ``"item_id"``, ``"target"``.
    ``entry["target"]`` is a ``np.ndarray`` with shape ``(num_ts, num_timestamps)``,
    and each row corresponds to one time series of the hierarchy.
    The goal is to reshape this DataEntry as a dataframe where:
    1) the index corresponds to the name of the time series,
    2) the columns ``"ds"`` and ``"y"`` correspond to timestamps and actuals, respectively.
    """

    df = pd.DataFrame(entry["target"]).T
    df.columns = S.index.tolist()
    df.index = pd.date_range(
        start=entry["start"].start_time,
        periods=df.shape[0],
        freq=entry["start"].freq,
    )

    df = unpivot(df)
    df = df.set_index("unique_id")

    return df


def unpivot(df: pd.DataFrame) -> pd.DataFrame:
    """
    Unpivot data frame.

    The input dataframe has as index the time stamps,
    and one column per each time series of the hierarchy.
    We unpivot this so that the final dataframe has
    three columns, i.e. ``"unique_id"``, ``"ds"``, and ``"y"``, where
    1) ``"unique_id"`` has the name of the corresponding time series,
    2) ``"ds"`` has the corresponding time stamps,
    3) ``"y"`` has the actuals.
    """

    n, k = df.shape
    return pd.DataFrame(
        {
            "unique_id": np.asarray(df.columns).repeat(n),
            "ds": np.tile(np.asarray(df.index), k),
            "y": df.to_numpy().ravel("F"),
        }
    )


def format_reconciled_forecasts(
    df: pd.DataFrame,
    prediction_length: int,
    fcst_col_name: str,
    S: pd.DataFrame,
):
    target_dim = S.shape[0]

    assert len(df) == target_dim * prediction_length

    # we want rows to correspond to time and columns to forecasts
    hier_forecasts = df.pivot(columns="ds", values=fcst_col_name)

    # we want columns to follow the order of the rows of S
    hier_forecasts = hier_forecasts.loc[S.index].T

    # we want rows to be sorted according to time
    hier_forecasts = hier_forecasts.sort_index()

    return np.array(hier_forecasts)


def prune_fcst_df(
    df: pd.DataFrame, base_reconciliation_model_name: str
) -> pd.DataFrame:
    # keep certain columns
    columns_to_keep = ["ds"]
    columns_to_keep.extend(
        [x for x in df.columns if base_reconciliation_model_name in x]
    )
    df = df[columns_to_keep]

    # rename columns
    mapper = {
        e: (
            "mean"
            if e == base_reconciliation_model_name
            else e.replace(f"{base_reconciliation_model_name}-", "")
        )
        for e in df.columns
    }
    df = df.rename(columns=mapper)

    return df


class HierarchicalForecastPredictor(RepresentablePredictor):
    """
    A predictor that wraps models from the `hierarchicalforecast`_ package.

    .. hierarchicalforecast: https://github.com/Nixtla/hierarchicalforecast

    Parameters
    ----------
    prediction_length
        Prediction length for the model to use.
    base_model
        forecaster to use for base forecasts. Please refer to the documentation
        of ``statsforecast`` for available options.
        Example: AutoARIMA
    reconciler
        forecast reconciliation method to use. Please see the documentation
        of ``hierarchicalforecast`` for available options.
        Example: BottomUp
    S
        Summation or aggregation matrix.
    tags
        Each key is a level with values of tags associated to that level
    ts_names
        ordered list with names of time series
    intervals_method
        method used to calculate prediction intervals.
        Options are `normality`, `bootstrap`, `permbu`.
    quantile_levels
        Optional list of quantile levels that we want predictions for.
        Note: this is only supported by specific types of models, such as
        ``AutoARIMA``. By default this is ``None``, giving only the mean
        prediction.
    n_jobs
        number of jobs used in the parallel processing, use -1 for all cores.
    model_params
        dictionary with inputs parameters for base_model,
        i.e. {"season_length": 2}
    reconciler_params
        dictionary with input parameters for reconciler,
        i.e. {"method": "average_proportions"}
    """

    @validated()
    def __init__(
        self,
        prediction_length: int,
        base_model: Any,
        reconciler: Any,
        S: Union[List[List[int]], np.ndarray],
        tags: Dict[str, Union[List, np.ndarray]] = {},
        ts_names: Optional[List[str]] = None,
        intervals_method: str = "normality",
        quantile_levels: Optional[List[float]] = None,
        n_jobs: int = 1,
        model_params: Dict[str, Any] = {},
        reconciler_params: Dict[str, Any] = {},
    ) -> None:
        super().__init__(prediction_length=prediction_length)

        assert intervals_method in ["normality", "bootstrap", "permbu"]

        if tags and ts_names:
            assert set([x for values in tags.values() for x in values]) == set(
                ts_names
            ), "tags and ts_names must have the same set of ts names"

        self.models = [base_model(**model_params)]
        self.hrec = HierarchicalReconciliation(
            reconcilers=[reconciler(**reconciler_params)]
        )
        self.ts_names = (
            ts_names
            if ts_names is not None
            else [str(x) for x in range(len(S))]
        )
        self.S = get_formatted_S(S, self.ts_names)
        self.tags = {key: np.array(val) for key, val in tags.items()}
        self.intervals_method = intervals_method
        self.n_jobs = n_jobs
        self.config = ModelConfig(
            quantile_levels=quantile_levels,
        )
        self.base_reconciliation_model_name = (
            f"{repr(self.models[0])}/"
            f"{_build_fn_name(self.hrec.reconcilers[0])}"
        )
        self.fitted = (
            repr(self.models[0]) not in models_without_fitted_capability
        )

    def predict_item(self, entry: DataEntry) -> QuantileForecast:
        kwargs = {}
        if self.config.intervals is not None and all(
            [
                proportion not in _build_fn_name(self.hrec.reconcilers[0])
                for proportion in [
                    "forecast_proportions",
                    "average_proportions",
                    "proportion_averages",
                ]
            ]
        ):
            kwargs["level"] = self.config.intervals

        Y_df = format_data_entry(entry, self.S)

        # set up forecaster
        forecaster = StatsForecast(
            df=Y_df,
            models=self.models,
            freq=entry["start"].freq,
            n_jobs=self.n_jobs,
        )

        # compute base forecasts
        Y_hat_df = forecaster.forecast(
            h=self.prediction_length, fitted=self.fitted, **kwargs
        )

        params = dict(
            Y_hat_df=Y_hat_df,
            S=self.S,
            tags=self.tags,
            intervals_method=self.intervals_method,
            **kwargs,
        )

        if self.fitted:
            params["Y_df"] = forecaster.forecast_fitted_values()
        else:
            params["Y_df"] = Y_df

        # reconcile forecasts
        Y_hat_df_rec = self.hrec.reconcile(**params)

        # select relevant columns and rename
        Y_hat_df_rec = prune_fcst_df(
            Y_hat_df_rec, self.base_reconciliation_model_name
        )

        # if only mean fcst is computed, we take it as all requested quantiles
        if len(Y_hat_df_rec.columns) == 2 and all(
            Y_hat_df_rec.columns == ["ds", "mean"]
        ):
            fcst_col_names = ["mean"] * len(self.config.statsforecast_keys)
        else:
            fcst_col_names = self.config.statsforecast_keys

        # prepare for QuantileForecast format
        forecast_arrays = np.array(
            [
                format_reconciled_forecasts(
                    df=Y_hat_df_rec,
                    fcst_col_name=fcst_col_names[e],
                    prediction_length=self.prediction_length,
                    S=self.S,
                )
                for e, k in enumerate(self.config.statsforecast_keys)
            ]
        )

        return QuantileForecast(
            forecast_arrays=forecast_arrays,
            forecast_keys=self.config.forecast_keys,
            start_date=forecast_start(entry),
            item_id=entry.get("item_id"),
            info=entry.get("info"),
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pydantic import BaseModel

from gluonts.core import serde
from gluonts.core.component import equals


class A(BaseModel):
    value: int


class B(BaseModel):
    a: A
    b: int


def test_nested_params():
    b = B(a=A(value=42), b=99)

    assert equals(b, serde.flat.decode(serde.flat.encode(b)))

    b2 = serde.flat.clone(b, {"a.value": 999})
    assert b2.a.value == 999


def test_kind():
    assert serde.flat.clone(A) is A
